"""Defines useful types for building your own YAML-compatible types."""

from enum import Enum
from inspect import isabstract
from typing import Dict, Tuple, Union

from .representers import register_int_like, register_str_like
from .yaml_lib import yaml

__all__ = ["YamlInt", "YamlIntEnum", "YamlStr", "YamlStrEnum"]


class YamlStr(str):
    """A `str` subclass that serializes to YAML as an integer."""

    def __init_subclass__(cls):
        """Adds a representer when defining a subclass."""
        res = super().__init_subclass__()
        if not isabstract(cls):
            register_str_like(cls, method=cls._to_yaml_str)
        return res

    # def __new__(cls, v):
    #     return super(cls).__new__(v)

    @classmethod
    def _to_yaml_str(cls, v) -> str:
        """Represent a value of this type as a string."""
        return str(v)

    def __str__(self) -> str:
        """This just returns the wrapped string."""
        return super().__str__()

    def __repr__(self) -> str:
        """Repr string that keeps the class type. You may redefine this."""
        return type(self).__qualname__ + "(" + super().__repr__() + ")"

    def __getnewargs__(self) -> Tuple[str]:
        """Implementation to prevent Enum from sabotaging us.
        
        I'm not joking! See `_make_class_unpicklable()` from `enum.py`
        """
        return super().__getnewargs__()


class YamlInt(int):
    """An `int` subclass that serializes to YAML as an integer."""

    def __init_subclass__(cls):
        """Adds a representer when defining a subclass."""
        res = super().__init_subclass__()
        if not isabstract(cls):
            register_int_like(cls, method=cls._to_yaml_int)
        return res

    # def __new__(cls, v):
    #     return super(cls).__new__(v)

    @classmethod
    def _to_yaml_int(cls, v) -> int:
        """Represent a value of this type as an integer."""
        return int(v)

    def __int__(self) -> int:
        return super().__int__()

    def __str__(self) -> str:
        """This returns the wrapped int."""
        return super().__str__()

    def __repr__(self) -> str:
        """Repr string that keeps the class type. You may redefine this."""
        return type(self).__qualname__ + "(" + super().__repr__() + ")"

    def __getnewargs__(self) -> Tuple[int]:
        """Implementation to prevent Enum from sabotaging us.
        
        I'm not joking! See `_make_class_unpicklable()` from `enum.py`
        """
        return super().__getnewargs__()


class YamlStrEnum(YamlStr, Enum):
    """String-valued enum that serializes to YAML directly as a string.

    This also checks that your enum can properly be serialized.
    """

    def __init_subclass__(cls):
        res = super().__init_subclass__()
        if not isabstract(cls):
            vals: Dict[str, cls] = dict(cls.__members__)
            yaml.safe_dump(vals)
        return res

    # def __new__(cls, v):
    #     return super(cls).__new__(v)

    def __str__(self) -> str:
        """This returns the wrapped string."""
        return str.__str__(self)

    def __repr__(self) -> str:
        """Repr string that keeps the class type. You may redefine this."""
        return type(self).__qualname__ + "(" + str.__repr__(self) + ")"


class YamlIntEnum(YamlInt, Enum):
    """Integer-valued enum that serializes to YAML directly as an integer.

    This also checks that your enum can properly be serialized.
    """

    def __init_subclass__(cls):
        res = super().__init_subclass__()
        if not isabstract(cls):
            vals: Dict[str, cls] = dict(cls.__members__)
            yaml.safe_dump(vals)
        return res


if __name__ == "__main__":

    class A(YamlStrEnum):
        a = "a"
        b = "b"

    class X(YamlIntEnum):
        x1 = 1
        x2 = 2



"""Basic tests for pydantic_yaml functionality."""


def test_readme():
    """Tests the functionality from the README."""


def test_simple():
    """Tests the functionality from the README."""

    from pydantic_yaml import YamlStrEnum, YamlModel

    class MyEnum(YamlStrEnum):
        a = "a"
        b = "b"

    class MyModel(YamlModel):
        x: int = 1
        e: MyEnum = MyEnum.a

    m1 = MyModel(x=2, e="b")
    yml = m1.yaml()
    jsn = m1.json()

    # This automatically assumes YAML
    m2 = MyModel.parse_raw(yml)
    assert m1 == m2

    # This explicitly sets YAML (or `content_type="application/yaml"`)
    m3 = MyModel.parse_raw(yml, proto="yaml")
    assert m1 == m3

    # This explicitly uses JSON
    m4 = MyModel.parse_raw(jsn, content_type="application/json")
    assert m1 == m4

    # JSON is actually a subset of YAML, so it should parse correctly anyways.
    m5 = MyModel.parse_raw(jsn)
    assert m1 == m5


def test_nested_models():
    """Test nested YAML models and inheritance."""

    from pydantic import BaseModel
    from pydantic_yaml import YamlModel, YamlModelMixin

    class A0(BaseModel):
        x: int = 1

    class A(A0, YamlModelMixin):
        x: int = 1

    class B(BaseModel):
        y: str = "hello"

    class C(YamlModel):
        a: A
        b: B

    c = C(a=A(x=11), b=B(y="bye"))

    yml = c.yaml()

    c_re = C.parse_raw(yml, proto="yaml")
    assert c == c_re



"""Module for some hacks that are needed to work around the YAML library interfaces.

Mainy, this defines "safe" representers for common types that Pydantic is able to parse.
"""

__all__ = ["inject_all"]

from typing import List, Type

from .representers import register_str_like, register_int_like


def get_str_like_types() -> List[Type]:
    """Returns many string-like types from stdlib and Pydantic."""

    # flake8: noqa

    from uuid import UUID
    from pathlib import (
        Path,
        PosixPath,
        PurePath,
        PurePosixPath,
        PureWindowsPath,
        WindowsPath,
    )

    from pydantic import types, networks

    def _chk(x) -> bool:
        try:
            return issubclass(x, (str, Path, PurePath, UUID))
        except Exception:
            return False

    candidates = (
        list(locals().values())
        + [getattr(types, v) for v in types.__all__]
        + [getattr(networks, v) for v in networks.__all__]
    )
    str_like = [v for v in candidates if _chk(v)]
    return str_like


def get_int_like_types() -> List[Type]:
    """Returns many int-like types from stdlib and Pydantic."""

    from pydantic import types

    def _chk(x) -> bool:
        try:
            return issubclass(x, (int))
        except Exception:
            return False

    candidates = list(locals().values()) + [getattr(types, v) for v in types.__all__]
    int_like = [v for v in candidates if _chk(v)]
    return int_like


def inject_all():
    """Injects all necessary "hacks" into the namespace.
    
    What this currently does:
    - Registers many str-like and int-like types (from Pydantic and the standard
      library) to be representable in YAML.
    """
    # Add representers for string-like and int-like values.
    for cls in get_str_like_types():
        register_str_like(cls, method=str)
    for cls in get_int_like_types():
        register_int_like(cls)



from typing import Optional, Tuple, Type
import warnings

from pydantic import validator

from pydantic_yaml.model import YamlModel
from .semver import SemVer

__all__ = ["VersionedYamlModel"]


def _chk_between(v, lo=None, hi=None):
    if v is None:
        return
    if (hi is not None) and (v > hi):
        raise ValueError(f"Default version higher than maximum: {v} > {hi}")
    if (lo is not None) and (v < lo):
        raise ValueError(f"Default version lower than minimum: {v} < {lo}")


def _get_minmax_robust(
    cls: Type["VersionedYamlModel"],
) -> Tuple[Optional[SemVer], Optional[SemVer]]:
    min_, max_ = None, None
    for supcls in cls.mro():
        Config = getattr(supcls, "Config", None)
        if Config is not None:
            if min_ is None:
                min_ = getattr(Config, "min_version", None)
            if max_ is None:
                max_ = getattr(Config, "max_version", None)
    return min_, max_


class VersionedYamlModel(YamlModel):
    """YAML model with versioning.

    Usage
    -----
    Inherit from this class, and set a Config class with attributes `min_version` and/or
    `max_version`:

    ```python
    class MyModel(VersionedYamlModel):
        class Config:
            min_version = 

    ```

    Notes
    -----
    By default, a validator checks that the "version" field is between 
    `Config.min_version` and `Config.max_version`, if those are not None.

    It's probably best not to even set the `version` field by hand.
    """

    version: SemVer

    def __init_subclass__(cls) -> None:
        # Check Config class types
        Config = getattr(cls, "Config", None)
        if Config is not None:
            # check one field
            minv = getattr(Config, "min_version", None)
            if minv is not None:
                if not isinstance(minv, SemVer):
                    setattr(Config, "min_version", SemVer(minv))
            # check other field
            maxv = getattr(Config, "max_version", None)
            if maxv is not None:
                if not isinstance(maxv, SemVer):
                    setattr(Config, "max_version", SemVer(maxv))

        # Check ranges
        min_, max_ = _get_minmax_robust(cls)
        if (min_ is not None) and (max_ is not None) and (min_ > max_):
            raise ValueError(
                f"Minimum version higher than maximum: {min_!r} > {max_!r}"
            )

        # Check the default value of the "version" field
        fld = cls.__fields__["version"]
        d = fld.default
        if d is None:
            pass
        else:
            _chk_between(d, lo=min_, hi=max_)
            warnings.warn(
                f"Recommended to have `version` be required, but set default {d!r}",
                UserWarning,
            )

        if not issubclass(fld.type_, SemVer):
            raise TypeError(
                f"Field type for `version` must be SemVer, got {fld.type_!r}"
            )

    @validator("version", always=True)
    def _check_semver(cls, v):
        min_, max_ = _get_minmax_robust(cls)
        _chk_between(v, lo=min_, hi=max_)
        return v

    class Config:
        min_version = SemVer("0.0.0")
        max_version = None



"""Module to define the YamlModelMixin."""

from typing_extensions import Literal
import warnings
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    no_type_check,
)
from pydantic.parse import Protocol, load_file, load_str_bytes
from pydantic.main import ROOT_KEY, BaseModel, ModelMetaclass, BaseConfig
from pydantic.error_wrappers import ErrorWrapper, ValidationError
from pydantic.types import StrBytes

if TYPE_CHECKING:
    from pydantic.typing import (
        DictStrAny,
        AbstractSetIntStr,
        MappingIntStrAny,
    )

    Model = TypeVar("Model", bound="BaseModel")

from .compat.yaml_lib import yaml


ExtendedProto = Union[Protocol, Literal["yaml"]]

YamlStyle = Union[
    None, Literal[""], Literal['"'], Literal["'"], Literal["|"], Literal[">"],
]


def is_yaml_requested(
    content_type: str = None,
    proto: ExtendedProto = None,
    path_suffix: Optional[str] = None,
) -> bool:
    """Checks whether YAML is requested by the user, depending on params."""
    is_yaml = False
    if content_type is not None:
        is_yaml = ("yaml" in content_type) or ("yml" in content_type)
    if proto is not None:
        is_yaml = is_yaml or (proto == "yaml")
    if path_suffix is not None:
        is_yaml = is_yaml or (path_suffix in [".yaml", ".yml"])
    return is_yaml


class YamlModelMixinConfig:
    """Additional configuration for YamlModelMixin."""

    yaml_loads: Callable[[str], Any] = yaml.safe_load  # type: ignore
    yaml_dumps: Callable[..., str] = yaml.safe_dump  # type: ignore


class YamlModelMixin(metaclass=ModelMetaclass):
    """Mixin to add YAML compatibility to your class.

    Usage
    -----
    Inherit from this and a `pydantic.BaseModel` or a subclass, like this:

    ```python
    class MyBaseType(BaseModel):
        my_field: str = "default"

    class MyExtType(YamlModelMixin, MyBaseType):
        other_field: int = 42
    ```

    `YamlModelMixin` *must* be *before* any class that inherits from
    `pydantic.BaseModel` due to issues with the method resolution order.

    You can now use `MyExtType().yaml()` to dump the class values (default here)
    to a YAML string.
    """

    if TYPE_CHECKING:
        __custom_root_type__: ClassVar[bool] = False
        __config__: ClassVar[Type[BaseConfig]] = YamlModelMixinConfig  # type: ignore

    @no_type_check
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        # Set the config class
        cfg: Type[YamlModelMixinConfig] = cls.__config__
        if not issubclass(cfg, YamlModelMixinConfig):

            class T(cfg, YamlModelMixinConfig):
                pass

            T.__doc__ = cfg.__doc__
            T.__name__ = cfg.__name__
            T.__qualname__ = cfg.__qualname__

            cls.__config__ = T

    def yaml(
        self,
        *,
        include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        default_flow_style: Optional[bool] = False,
        default_style: YamlStyle = None,
        indent: Optional[bool] = None,
        encoding: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Generate a YAML representation of the model.

        Parameters
        ----------
        include, exclude
            Fields to include or exclude. See `dict()`.
        by_alias : bool
            Whether to use aliases instead of declared names. Default is False.
        skip_defaults, exclude_unset, exclude_defaults, exclude_none
            Arguments as per `BaseModel.dict()`.
        default_flow_style : bool or None
            Whether to use the "flow" style in the dumper. By default, this is False,
            which uses the "block" style (probably the most familiar to users).
        default_style : {None, "", "'", '"', "|", ">"}
            This is the default style for quoting strings, used by `ruamel.yaml` dumper.
            Default is None, which varies the style based on line length.
        indent, encoding, kwargs
            Additional arguments for the dumper.
        """
        data: "DictStrAny" = self.dict(  # type: ignore
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        if self.__custom_root_type__:
            data = data[ROOT_KEY]
            warnings.warn(
                "Explicit custom root behavior not yet implemented for pydantic_yaml."
                " This may not work as expected. If so, please create a GitHub issue:"
                " https://github.com/NowanIlfideme/pydantic-yaml/issues/new"
            )
        cfg = cast(YamlModelMixinConfig, self.__config__)
        return cfg.yaml_dumps(
            data,
            default_flow_style=default_flow_style,
            default_style=default_style,
            encoding=encoding,
            indent=indent,
            **kwargs,
        )

    @no_type_check
    @classmethod
    def parse_raw(
        cls: Type["Model"],
        b: StrBytes,
        *,
        content_type: str = "application/yaml",  # This is a reasonable default, right?
        encoding: str = "utf-8",
        proto: ExtendedProto = None,
        allow_pickle: bool = False,
    ) -> "Model":
        # NOTE: Type checking this function is a PITA, because we're overriding
        # BaseModel.parse_raw, but not inheriting it due to the MRO problem!

        # Check whether we're specifically asked to parse YAML
        is_yaml = is_yaml_requested(content_type=content_type, proto=proto)

        # Note that JSON is a subset of the YAML 1.2 spec, so we should be OK
        # even if JSON is passed. It will be slower, however.
        if is_yaml:
            try:
                cfg = cast(YamlModelMixinConfig, cls.__config__)
                obj = cfg.yaml_loads(b)
            except RecursionError as e:
                raise ValueError(
                    "YAML files with recursive references are unsupported."
                ) from e
            except ValidationError:
                raise
            except Exception as e:
                raise ValidationError([ErrorWrapper(e, loc=ROOT_KEY)], cls) from e
        else:
            obj = load_str_bytes(
                b,
                proto=proto,
                content_type=content_type,
                encoding=encoding,
                allow_pickle=allow_pickle,
                json_loads=cls.__config__.json_loads,
            )
        res = cls.parse_obj(obj)  # type: ignore
        return cast("Model", res)

    @no_type_check
    @classmethod
    def parse_file(
        cls: Type["Model"],
        path: Union[str, Path],
        *,
        content_type: str = None,
        encoding: str = "utf-8",
        proto: ExtendedProto = None,
        allow_pickle: bool = False,
    ) -> "Model":
        path = Path(path)

        # Assume YAML based on the file name, so `parse_raw()` works well below
        if (content_type is None) and (path.suffix in [".yml", ".yaml"]):
            content_type = "application/yaml"

        # Check whether we're specifically asked to parse YAML
        is_yaml = is_yaml_requested(
            content_type=content_type, proto=proto, path_suffix=path.suffix
        )

        # The first code path explicitly checks YAML compatibility.
        # We offload the rest to Pydantic.
        if is_yaml:
            b = path.read_bytes()
            return cls.parse_raw(
                b,
                content_type=content_type,
                encoding=encoding,
                proto=proto,
                allow_pickle=allow_pickle,
            )
        else:
            obj = load_file(
                path,
                proto=proto,
                content_type=content_type,
                encoding=encoding,
                allow_pickle=allow_pickle,
                json_loads=cls.__config__.json_loads,
            )
            res = cls.parse_obj(obj)  # type: ignore
            return cast("Model", res)



"""Tests for file IO."""

# type: ignore

from json.decoder import JSONDecodeError
from pathlib import Path
from typing import no_type_check

import pytest


@no_type_check
def test_file_io(tmpdir: str):
    """Test file IO with temporary files (tmpdir is a pytest fixture)."""

    from pydantic import BaseModel
    from pydantic_yaml import YamlModel, YamlModelMixin, YamlStrEnum

    class MyEnum(YamlStrEnum):
        a = "a"
        b = "b"

    class MyModel(YamlModel):
        x: int = 1
        e: MyEnum = MyEnum.a

    class TestBase(BaseModel):
        tag: str = "my tag"

    class MyOtherModel(YamlModelMixin, TestBase):
        sub: MyModel

    m1 = MyModel(x=2, e="b")
    m2 = MyOtherModel(sub=m1)

    base_dir: Path = Path(tmpdir).resolve()

    for (m, M) in zip([m1, m2], [MyModel, MyOtherModel]):
        with (base_dir / "mdl.yaml").open(mode="w") as f:
            f.write(m.yaml())
        with (base_dir / "mdl.json").open(mode="w") as f:
            f.write(m.json())
        assert M.parse_file(base_dir / "mdl.yaml") == m
        assert M.parse_file(base_dir / "mdl.json") == m
        assert M.parse_file(base_dir / "mdl.yaml", content_type=None) == m
        assert M.parse_file(base_dir / "mdl.json", content_type=None) == m

        with pytest.raises(JSONDecodeError):
            M.parse_file(base_dir / "mdl.yaml", content_type="application/json")
        # This works because JSON is a subset of YAML :)
        mx = M.parse_file(base_dir / "mdl.json", content_type="application/yaml")
        assert mx == m



"""Old class for YamlEnum that is now deprecated (and marked as such)."""

from enum import Enum
from inspect import isabstract

from deprecated import deprecated

from .representers import register_str_like

__all__ = ["YamlEnum"]


@deprecated(version="0.5.0", reason="Use the `YamlStrEnum` class instead.")
class YamlEnum(Enum):
    """This class is DEPRECATED, please use `pydantic_yaml.YamlStrEnum` instead.
    
    Enumeration that serializes as the proper underlying object type.

    You can use this instead of `enum.Enum`, for example:
        class MyEnum(str, YamlEnum):
            val1 = "Value 1"
            val2 = "Value 2"
    
    Note
    ----
    This is actually a lie, it only supports `str` enums for now. Oops.
    """

    def __init_subclass__(cls):
        if not isabstract(cls):
            register_str_like(cls)

    def __repr__(self) -> str:
        return repr(self.value)

    def __str__(self) -> str:
        return str(self.value)



"""Imports an installed YAML library."""

# flake8: noqa

try:
    import ruamel.yaml as yaml  # type: ignore

    # ruamel.yaml doesn't have type annotations

    if yaml.__version__ < "0.15.0":
        __yaml_lib__ = "ruamel-old"
    else:
        __yaml_lib__ = "ruamel-new"
except ImportError:
    try:
        import yaml  # type: ignore

        __yaml_lib__ = "pyyaml"
    except ImportError:
        raise ImportError(
            "Could not import ruamel.yaml or pyyaml, "
            "please install at least one of them."
        )

dumper_classes = []
for _fld in dir(yaml):
    try:
        if "Dumper" in _fld:
            _obj = getattr(yaml, _fld)
            dumper_classes.append(_obj)
    except Exception:
        pass


__all__ = ["yaml", "__yaml_lib__"]



"""Representers for dumping common objects to YAML."""

from functools import partial
from typing import Any, Callable, TypeVar

from .yaml_lib import yaml, dumper_classes

CType = TypeVar("CType")

__all__ = ["register_str_like", "register_int_like"]


def dump_as_str(dumper: yaml.Dumper, data: Any, method: Callable[[Any], str] = str):
    """Represents an object as a string in YAML.

    Parameters
    ----------
    dumper : yaml.Dumper
        The dumper instance, such as `yaml.SafeDumper()`.
    data
        The object to dump. Ideally this is string-like.
    method : callable
        A method that converts `data` to a string. Default is `str`.

    Returns
    -------
    node
        The YAML internal node representation.
    """
    return dumper.represent_str(method(data))


def dump_as_int(dumper: yaml.Dumper, data: Any, method: Callable[[Any], int] = int):
    """Represents an object as an integer in YAML.

    Parameters
    ----------
    dumper : yaml.Dumper
        The dumper instance, such as `yaml.SafeDumper()`.
    data
        The object to dump. Ideally this is int-like.
    method : callable
        A method that converts `data` to an integer. Default is `int`.

    Returns
    -------
    node
        The YAML internal node representation.
    """
    return dumper.represent_int(method(data))


def register_str_like(cls: CType, method: Callable[[Any], str] = str) -> CType:
    """Registers `cls` to be dumped to YAML as a string.

    Parameters
    ----------
    cls : Type
        The class to represent.
    method : callable
        A method that converts objects of type `cls` to a string.
        Default is `str`.

    Returns
    -------
    cls
        This is the same as the input `cls`.
    """
    for dump_cls in dumper_classes:
        dump_cls.add_representer(cls, partial(dump_as_str, method=method))
    return cls


def register_int_like(cls: CType, method: Callable[[Any], int] = int) -> CType:
    """Registers `cls` to be dumped to YAML as an integer.

    Parameters
    ----------
    cls : Type
        The class to represent.
    method : callable
        A method that converts objects of type `cls` to an integer.
        Default is `int`.

    Returns
    -------
    cls
        This is the same as the input `cls`.
    """
    for dump_cls in dumper_classes:
        dump_cls.add_representer(cls, partial(dump_as_int, method=method))
    return cls
